#===============================================================================
# index.py
#===============================================================================

"""Index dbSNP VCFs by rsid"""



# Imports ======================================================================

import os.path
import subprocess

from argparse import ArgumentParser
from functools import partial
from multiprocessing import Pool

from pydbsnp.env import VCF_GRCH37, VCF_GRCH38, RSID_GRCH37, RSID_GRCH38




# Functions ====================================================================

def reformat_sort(
    input_vcf_path,
    output_rsid_path,
    quiet=False,
    temp_dir=None
):
    if not quiet:
        print(
            f'Reformatting database {input_vcf_path} and sorting by RSID. '
            'This will take a while. Reformatted and sorted data will be '
            f'written to {output_rsid_path}.'
        )
    with open(output_rsid_path, 'wb') as f:
        with subprocess.Popen(
            ('zcat', input_vcf_path), stdout=subprocess.PIPE
        ) as zcat:
            with subprocess.Popen(
                (
                    'awk', '-v', 'OFS=\t',
                    (
                        '!/##/ && !/#CHROM/ '
                        '{sub(/rs/, "", $3); print "rs", $3, $1, $3}'
                    ),
                ),
                stdin=zcat.stdout,
                stdout=subprocess.PIPE
            ) as awk:
                with subprocess.Popen(
                    (
                        'sort',
                        '-k2,2',
                        '-n',
                        '-T', temp_dir or os.path.dirname(output_rsid_path)
                    ),
                    stdin=awk.stdout,
                    stdout=subprocess.PIPE
                ) as sort:
                    subprocess.run(
                        ('bgzip', '--stdout'),
                        stdin=sort.stdout,
                        stdout=f
                    )
                

def index(rsid_file_path, quiet=False):
    if not quiet:
        print(f'Indexing {rsid_file_path}.')
    subprocess.run(
        (
            'tabix',
            '--csi',
            '--sequence', '1',
            '--begin', '2',
            '--end', '2', 
            rsid_file_path
        )
    )


def reformat_sort_index(
    input_vcf_path,
    output_rsid_path,
    quiet=False,
    temp_dir=None
):
    reformat_sort(
        input_vcf_path,
        output_rsid_path,
        quiet=quiet,
        temp_dir=temp_dir
    )
    index(output_rsid_path, quiet=False)
    

def parse_arguments():
    parser = ArgumentParser(description='index dbSNP VCF data by rsid')
    parser.add_argument(
        '--quiet',
        action='store_true',
        help='suppress printed status updates'
    )
    parser.add_argument(
        '--processes',
        choices=(1, 2),
        default=1,
        help='set to 2 to index GRCh37 and GRCh38 data in parallel'
    )
    parser.add_argument(
        '--tmp-dir',
        metavar='<path/to/tmp/dir/>',
        help='directory for temporary files'
    )
    return parser.parse_args()


def main():
    args = parse_arguments()
    with Pool(processes=args.processes) as pool:
        pool.starmap(
            partial(
                reformat_sort_index,
                quiet=args.quiet,
                temp_dir=args.tmp_dir
            ),
            (
                (vcf, rsid)
                for vcf, rsid
                in ((VCF_GRCH37, RSID_GRCH37), (VCF_GRCH38, RSID_GRCH38))
                if os.path.isfile(vcf)
            )
        )



#===============================================================================
# download.py
#===============================================================================

"""Download VCF data from dbSNP"""



# Imports ======================================================================

from argparse import ArgumentParser
from ftplib import FTP

from pydbsnp.env import FTP_BASENAME_GRCH37, FTP_BASENAME_GRCH38, BUILD_TO_VCF




# Constants ====================================================================

FTP_HOST = 'ftp.ncbi.nlm.nih.gov'
FTP_DIR = 'snp/latest_release/VCF'
BUILD_TO_FTP_BASENAME = {
    'hg19': FTP_BASENAME_GRCH37, 'GRCh37': FTP_BASENAME_GRCH37,
    'hg38': FTP_BASENAME_GRCH38, 'GRCh38': FTP_BASENAME_GRCH38
}
BUILD_TO_INT = {'hg19': 37, 'GRCh37':37, 'hg38': 38,'GRCh38': 38}




# Functions ====================================================================

def download(reference_build='GRCh38', quiet=False):
    if not quiet:
        print(
            'Downloading dbSNP data in VCF format '
            f'({reference_build} coordinates) to '
            f'{BUILD_TO_VCF[reference_build]}. '
            'This will probably take a few minutes.'
        )
    ftp = FTP(FTP_HOST)
    ftp.login()
    ftp.cwd(FTP_DIR)
    with open(BUILD_TO_VCF[reference_build], 'wb') as f:
        ftp.retrbinary(
            f'RETR {BUILD_TO_FTP_BASENAME[reference_build]}', f.write
        )
    if not quiet:
        print(
            f'Downloading tabix index to {BUILD_TO_VCF[reference_build]}.tbi.'
        )
    with open(f'{BUILD_TO_VCF[reference_build]}.tbi', 'wb') as f:
        ftp.retrbinary(
            f'RETR {BUILD_TO_FTP_BASENAME[reference_build]}.tbi', f.write
        )
    ftp.quit()
    if not quiet:
        print(f'Download complete.')


def parse_arguments():
    parser = ArgumentParser(description='download dbSNP VCF data')
    parser.add_argument(
        '--reference-build',
        choices=('hg19', 'GRCh37', 'hg38', 'GRCh38'),
        default='GRCh38',
        help='reference genome build'
    )
    parser.add_argument(
        '--quiet',
        action='store_true',
        help='suppress printed status updates'
    )
    return parser.parse_args()


def main():
    args = parse_arguments()
    download(reference_build=args.reference_build, quiet=args.quiet)



#===============================================================================
# query.py
#===============================================================================

"""query the dbSNP VCF data"""




# Imports ======================================================================

import re

from argparse import ArgumentParser
from pysam import TabixFile, VariantFile

from pydbsnp.env import BUILD_TO_VCF, BUILD_TO_RSID




# Constants ====================================================================

COORD_REGEX = re.compile('.+:[0-9]+$')
RSID_REGEX = re.compile('rs[1-9][0-9]+$')




# Functions ====================================================================

def parse_arguments():
    parser = ArgumentParser(description='query dbSNP VCF data')
    parser.add_argument(
        'variants',
        nargs='+',
        metavar='<rsid or chr:pos>',
        help='variant for which to query database'
    )
    parser.add_argument(
        '-r',
        '--reference-build',
        choices=('GRCh37', 'hg19', 'GRCh38', 'hg38'),
        default='GRCh38',
        help='reference build for coordinates'
    )
    return parser.parse_args()


def main():
    args = parse_arguments()
    print(VariantFile(BUILD_TO_VCF[args.reference_build]).header)
    vcf = TabixFile(BUILD_TO_VCF[args.reference_build])
    rsid = TabixFile(BUILD_TO_RSID[args.reference_build])
    def rsid_to_coordinates(rsid, variant):
        rs_number = variant.replace('rs', '')
        for row in rsid.fetch('rs', rs_number):
            yield row[2], int(row[3])
    for variant in args.variants:
        if COORD_REGEX.match(variant):
            chrom, pos = variant.split(':')
            pos = int(pos)
            for row in vcf.fetch(chrom, pos, pos):
                print(str(row))
        elif RSID_REGEX.match(variant):
            for chrom, pos in rsid_to_coordinates(rsid, variant):
                for row in vcf.fetch(chrom, pos, pos):
                    print(str(row))
        else:
            raise RuntimeError('Improperly formatted query')
